/*

REALS.cc -- main part of the implementation of reals for the iRRAM library
 
Copyright (C) 2001-2003 Norbert Mueller
 
This file is part of the iRRAM Library.
 
The iRRAM Library is free software; you can redistribute it and/or modify
it under the terms of the GNU Library General Public License as published by
the Free Software Foundation; either version 2 of the License, or (at your
option) any later version.
 
The iRRAM Library is distributed in the hope that it will be useful, but
WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANTABILITY
or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU Library General Public
License for more details.
 
You should have received a copy of the GNU Library General Public License
along with the iRRAM Library; see the file COPYING.LIB.  If not, write to
the Free Software Foundation, Inc., 59 Temple Place - Suite 330, Boston,
MA 02111-1307, USA. 
*/

#include <stdlib.h>
#include <stdarg.h>
#include <string.h>

#include "iRRAM_core.h"

#if defined(LRGMP)
	#include "LRGMP_ext.h"
	#define iRRAM_BACKEND "LRGMP"
#elif defined(cGMP)
	#include "cGMP_ext.h"
	#define iRRAM_BACKEND "cGMP"
#elif defined(GMP)
	#include "GMP_ext.h"
	#define iRRAM_BACKEND "GMP"
#elif defined(MPFR)
	#include "MPFR_ext.h"
	#define iRRAM_BACKEND "MPFR"
#else
	#include "Backend not defined"
#endif


REAL::REAL(){
  MP_init(value); 
  MP_int_to_mp(0,value); 
  sizetype_exact(error);
  MP_getsize(value,vsize);
};

REAL::REAL(const int i){
  MP_init(value);
  MP_int_to_mp(i,value);
  sizetype_exact(error);
  MP_getsize(value,vsize);
};

REAL::REAL(const DYADIC& y){ 
  MP_duplicate_w_init(y.value,value);
  sizetype_exact(error);
  MP_getsize(value,vsize);
};

REAL::REAL(const REAL& y){
  MP_duplicate_w_init(y.value,value);
  error=y.error;
  vsize=y.vsize;
};

REAL::REAL(const string s){
  REAL result=atoREAL(s.c_str());
  error=result.error;
  MP_duplicate_w_init(result.value,value);
  MP_getsize(value,vsize);
};

REAL::REAL(const char* s){
  REAL result=atoREAL(s);
  error=result.error;
  MP_duplicate_w_init(result.value,value);
  MP_getsize(value,vsize);
};


REAL operator ^ (const REAL& x, const REAL& y) {
  return power(x,y);
}

REAL operator ^ (const REAL& x, int n) {
  return power(x,n);
}

REAL operator << (const REAL& x, int n) {
  return scale(x,n);
}

REAL operator >> (const REAL& x, int n) {
  return scale(x,-n);
}

REAL operator + (const REAL& x, const REAL& y) {
  MP_type zvalue; sizetype zerror; int local_prec;
  if (ACTUAL_STACK.prec_policy == 0)
  local_prec=max3(y.error.exponent,x.error.exponent,ACTUAL_STACK.actual_prec);
  else {
  local_prec=max(x.vsize.exponent,y.vsize.exponent);
  local_prec=max3(y.error.exponent,x.error.exponent,local_prec-50+ACTUAL_STACK.actual_prec);
  }
  MP_init(zvalue);
  MP_mv_add(x.value,y.value,zvalue,local_prec);
  sizetype_set(zerror,1,local_prec);
  sizetype_inc(zerror,x.error);
  sizetype_inc(zerror,y.error);
  ADJUST(zerror);
  return REAL(zvalue,zerror);
};

REAL operator + (const REAL& x, int n)
{
  MP_type zvalue; sizetype zerror; int local_prec=0;
  if (ACTUAL_STACK.prec_policy == 0) 
    local_prec=max(x.error.exponent,ACTUAL_STACK.actual_prec);
  else {
    sizetype ysize;
    sizetype_set(ysize,n>0?n:-n,0);
    local_prec=max(x.vsize.exponent,ysize.exponent);
    local_prec=max(x.error.exponent,local_prec-50+ACTUAL_STACK.actual_prec);
  }
  MP_init(zvalue);
  MP_mv_addi(x.value,n,zvalue,local_prec);
  sizetype_set(zerror,1,local_prec);
  sizetype_inc(zerror,x.error);
  ADJUST(zerror);
  return REAL(zvalue,zerror);
}

REAL operator + (int n, const REAL& x)
{
  return (x+n);
}

REAL& operator += (REAL& x,const REAL& y) {
  MP_type zvalue;
  sizetype zerror; int local_prec;
  if (ACTUAL_STACK.prec_policy == 0)
  local_prec=max3(y.error.exponent,x.error.exponent,ACTUAL_STACK.actual_prec);
  else {
  local_prec=max(x.vsize.exponent,y.vsize.exponent);
  local_prec=max3(y.error.exponent,x.error.exponent,local_prec-50+ACTUAL_STACK.actual_prec);
  }
  MP_init(zvalue);
  MP_mv_add(x.value,y.value,zvalue,local_prec);
  sizetype_set(zerror,1,local_prec);
  sizetype_inc(x.error,y.error);
  sizetype_inc(x.error,zerror);
  ADJUST(x.error);
  MP_clear(x.value);
  x.value=zvalue;
  MP_getsize(x.value,x.vsize);
  return x;
};

REAL operator - (const REAL& x, const REAL& y) {
  MP_type zvalue; sizetype zerror; int local_prec;
  if (ACTUAL_STACK.prec_policy == 0)
  local_prec=max3(y.error.exponent,x.error.exponent,ACTUAL_STACK.actual_prec);
  else
  {
  local_prec=max(x.vsize.exponent,y.vsize.exponent);
  local_prec=max3(y.error.exponent,x.error.exponent,local_prec-50+ACTUAL_STACK.actual_prec);
  }
  MP_init(zvalue);
  MP_mv_sub(x.value,y.value,zvalue,local_prec);
  sizetype_set(zerror,1,local_prec);
  sizetype_inc(zerror,x.error);
  sizetype_inc(zerror,y.error);
  ADJUST(zerror);
  return REAL(zvalue,zerror);
};

REAL operator - (const REAL& x, int n) 
{
  MP_type zvalue; 
  sizetype zerror; 
  int local_prec;
  if (ACTUAL_STACK.prec_policy == 0)
    local_prec=max(x.error.exponent,ACTUAL_STACK.actual_prec);
  else {
    sizetype ysize;
    sizetype_set(ysize,n>0?n:-n,0);
    local_prec=max(x.vsize.exponent,ysize.exponent);
    local_prec=max(x.error.exponent,local_prec-50+ACTUAL_STACK.actual_prec);
  }
  MP_init(zvalue);
  MP_mv_subi(x.value,n,zvalue,local_prec);
  sizetype_set(zerror,1,local_prec);
  sizetype_inc(zerror,x.error);
  ADJUST(zerror);
  return REAL(zvalue,zerror);
};


REAL operator - (int n,const REAL& x) 
{
  	MP_type zvalue;
	sizetype zerror;
	int local_prec;
  	if (ACTUAL_STACK.prec_policy == 0)
  		local_prec=max(x.error.exponent,ACTUAL_STACK.actual_prec);
  	else
  		{
  		sizetype xsize,ysize;
  		MP_getsize(x.value,xsize);
		sizetype_set(ysize,n>0?n:-n,0);
		local_prec=max(xsize.exponent,ysize.exponent);
  		local_prec=max(x.error.exponent,local_prec-50+ACTUAL_STACK.actual_prec);
  		}
  	MP_init(zvalue);
  	MP_mv_isub(n,x.value,zvalue,local_prec);
  	sizetype_set(zerror,1,local_prec);
  	sizetype_inc(zerror,x.error);
  	ADJUST(zerror);
  	return REAL(zvalue,zerror);
};


REAL operator - (const REAL& x) {
  REAL y=REAL(0);
  MP_type zvalue; sizetype zerror; int local_prec;
  if (ACTUAL_STACK.prec_policy == 0)
  local_prec=max3(y.error.exponent,x.error.exponent,ACTUAL_STACK.actual_prec);
  else
  {
  local_prec=max(x.vsize.exponent,y.vsize.exponent);
  local_prec=max3(y.error.exponent,x.error.exponent,local_prec-50+ACTUAL_STACK.actual_prec);
  }
  MP_init(zvalue);
  MP_mv_sub(y.value,x.value,zvalue,local_prec);
  sizetype_set(zerror,1,local_prec);
  sizetype_inc(zerror,x.error);
  sizetype_inc(zerror,y.error);
  ADJUST(zerror);
  return REAL(zvalue,zerror);
};

REAL operator * (const REAL& x, const REAL& y) {
  MP_type zvalue;
  sizetype zerror,proderror;
  int local_prec;
  sizetype_mult(zerror,x.vsize,y.error);
  sizetype_mult(proderror,y.vsize,x.error);
  sizetype_inc(zerror,proderror);
  sizetype_mult(proderror,y.error,x.error);
  sizetype_inc(zerror,proderror);
  if (ACTUAL_STACK.prec_policy == 0)
    local_prec=max(zerror.exponent,ACTUAL_STACK.actual_prec);
  else
    local_prec=max(zerror.exponent,x.vsize.exponent+y.vsize.exponent-50+ACTUAL_STACK.actual_prec);
  MP_init(zvalue);
  MP_mv_mul(x.value,y.value,zvalue,local_prec);
  sizetype_set(proderror,1,local_prec);
  sizetype_inc(zerror,proderror);
  ADJUST(zerror);
  return REAL(zvalue,zerror);
};

REAL& operator *= (REAL& x,const REAL& y) {
  MP_type zvalue;
  sizetype zerror,proderror;
  int local_prec;
  sizetype_mult(zerror,x.vsize,y.error);
  sizetype_mult(proderror,y.vsize,x.error);
  sizetype_inc(zerror,proderror);
  sizetype_mult(proderror,y.error,x.error);
  sizetype_inc(zerror,proderror);
  if (ACTUAL_STACK.prec_policy == 0)
    local_prec=max(zerror.exponent,ACTUAL_STACK.actual_prec);
  else
    local_prec=max(zerror.exponent,x.vsize.exponent+y.vsize.exponent-50+ACTUAL_STACK.actual_prec);
  MP_init(zvalue);
  MP_mv_mul(x.value,y.value,zvalue,local_prec);
  sizetype_set(proderror,1,local_prec);
  sizetype_inc(zerror,proderror);
  ADJUST(zerror);
  x.error=zerror;
  MP_clear(x.value);
  x.value=zvalue;
  MP_getsize(x.value,x.vsize);
  return x;
};

REAL operator * (const REAL& x, int n) 
{
  	MP_type zvalue;
  	sizetype zerror,proderror,ysize;
  	int local_prec;
	sizetype_set(ysize,n>0?n:-n,0);
  	sizetype_mult(zerror,ysize,x.error);

  	if (ACTUAL_STACK.prec_policy == 0)
    		local_prec=max(zerror.exponent,ACTUAL_STACK.actual_prec);
  	else
    		local_prec=max(zerror.exponent,x.vsize.exponent+ysize.exponent-50+ACTUAL_STACK.actual_prec);
  	MP_init(zvalue);
  	MP_mv_muli(x.value,n,zvalue,local_prec);
  	sizetype_set(proderror,1,local_prec);
  	sizetype_inc(zerror,proderror);
  	ADJUST(zerror);
  	return REAL(zvalue,zerror);
}

REAL operator * (int n, const REAL& x){
  return (x*n);
}

REAL& operator *= (REAL& x,int n) 
{	
  MP_type zvalue;
  sizetype zerror,proderror,ysize;
  int local_prec;
  sizetype_set(ysize,n>0?n:-n,0);
  sizetype_mult(zerror,ysize,x.error);
  if (ACTUAL_STACK.prec_policy == 0)
    	local_prec=max(zerror.exponent,ACTUAL_STACK.actual_prec);
  else
    	local_prec=max(zerror.exponent,x.vsize.exponent+ysize.exponent-50+ACTUAL_STACK.actual_prec);
  MP_init(zvalue);
  MP_mv_muli(x.value,n,zvalue,local_prec);
  sizetype_set(proderror,1,local_prec);
  sizetype_inc(zerror,proderror);
  ADJUST(zerror);
  x.error=zerror;
  MP_clear(x.value);
  x.value=zvalue;
  MP_getsize(x.value,x.vsize);
  return x;

}

REAL operator / (const REAL& x, const REAL& y) {
  MP_type zvalue; 
  sizetype zerror,h1,h2,h3;
  int local_prec;
  sizetype_half(h1,y.vsize);
  if (sizetype_less(h1,y.error)) {
      DEBUG5(1,"insufficient precision %d*2^(%d) in denominator of size %d*2^(%d)\n",
              y.error.mantissa,y.error.exponent,y.vsize.mantissa,y.vsize.exponent);
      REITERATE(0);
    }
  sizetype_mult(h1,x.vsize,y.error);
  sizetype_mult(h2,y.vsize,x.error);
  sizetype_inc(h1,h2);
  h3=y.vsize;
  sizetype_dec(h3);
  sizetype_dec(h3,y.error);
  sizetype_mult(h2,h3,y.vsize);
  sizetype_div(zerror,h1,h2);
  if (ACTUAL_STACK.prec_policy == 0)
    local_prec=max(zerror.exponent,ACTUAL_STACK.actual_prec);
  else
    local_prec=max(zerror.exponent,x.vsize.exponent-y.vsize.exponent-50+ACTUAL_STACK.actual_prec);
  MP_init(zvalue);
  MP_mv_div(x.value,y.value,zvalue,local_prec);
  sizetype_set(h2,1,local_prec);
  sizetype_inc(zerror,h2);
  ADJUST(zerror);
  return REAL(zvalue,zerror);
};

REAL operator / (const REAL& x, int n) {
  	MP_type zvalue; 
  	sizetype zerror,ysize,h1;
  	int local_prec;
	sizetype_set(ysize,n>0?n:-n,0);
  	sizetype_div(zerror,x.error,ysize);
  	if (ACTUAL_STACK.prec_policy == 0)
    		local_prec=max(zerror.exponent,ACTUAL_STACK.actual_prec);
  	else
    		local_prec=max(zerror.exponent,x.vsize.exponent-ysize.exponent-50+ACTUAL_STACK.actual_prec);
  	MP_init(zvalue);
  	MP_mv_divi(x.value,n,zvalue,local_prec);
  	sizetype_set(h1,1,local_prec);
  	sizetype_inc(zerror,h1);
  	ADJUST(zerror);
  	return REAL(zvalue,zerror);
}

REAL operator / (int n, const REAL& y) {
  	MP_type zvalue; 
  	sizetype zerror,xsize,h1,h2,h3,xerror;
  	int local_prec;
	sizetype_exact(xerror);
  	sizetype_set(xsize,n>0?n:-n,0);
  	sizetype_half(h1,y.vsize);

  	if (sizetype_less(h1,y.error)) 
		{
      	DEBUG5(1,"insufficient precision %d*2^(%d) in denominator of size%d*2^(%d)\n",
	           y.error.mantissa,y.error.exponent,y.vsize.mantissa,y.vsize.exponent);
      	REITERATE(0);
    		}
  	sizetype_mult(h1,xsize,y.error);
  	sizetype_mult(h2,y.vsize,xerror);
  	sizetype_inc(h1,h2);
  	h3=y.vsize;
  	sizetype_dec(h3);
  	sizetype_dec(h3,y.error);
  	sizetype_mult(h2,h3,y.vsize);
  	sizetype_div(zerror,h1,h2);
  	if (ACTUAL_STACK.prec_policy == 0)
    		local_prec=max(zerror.exponent,ACTUAL_STACK.actual_prec);
  	else
    		local_prec=max(zerror.exponent,xsize.exponent-y.vsize.exponent-50+ACTUAL_STACK.actual_prec);
  	MP_init(zvalue);
  	MP_mv_idiv(n,y.value,zvalue,local_prec);
  	sizetype_set(h2,1,local_prec);
  	sizetype_inc(zerror,h2);
  	ADJUST(zerror);

  	return REAL(zvalue,zerror);
}



REAL square (const REAL& x) {
  MP_type zvalue;
  sizetype zerror,proderror;
  int local_prec;
  sizetype_mult(zerror,x.vsize,x.error);
  sizetype_mult(proderror,x.vsize,x.error);
  sizetype_inc(zerror,proderror);
  sizetype_mult(proderror,x.error,x.error);
  sizetype_inc(zerror,proderror);
  if (ACTUAL_STACK.prec_policy == 0)
    local_prec=max(zerror.exponent,ACTUAL_STACK.actual_prec);
  else
    local_prec=max(zerror.exponent,x.vsize.exponent+x.vsize.exponent-50+ACTUAL_STACK.actual_prec);
  MP_init(zvalue);
  MP_mv_mul(x.value,x.value,zvalue,local_prec);
  sizetype_set(proderror,1,local_prec);
  sizetype_inc(zerror,proderror);
  ADJUST(zerror);
  return REAL(zvalue,zerror);
};


#ifdef MP_shift

REAL scale(const REAL& x, int n) {
  sizetype zerror;
  MP_type zvalue;
  MP_init(zvalue);
  MP_shift(x.value,zvalue,n);
  x.geterror(zerror);
  sizetype_shift(zerror,zerror,n);
  return REAL(zvalue,zerror);
}

#else

REAL scale(const REAL& x, int n) {
   if (n==0) return x;
   if (n==1) return 2*x;
   stiff_begin();
   REAL y=1;
   REAL xc=2;
   if (n<0) {xc=y/2;n=-n;}
   if (n==1) {
     stiff_end();
     return xc*x;
   }
   for (int k=n;k>0;k=k/2) { 
     if (k%2==1) y*=xc;
     if ( k ==1) break;
     xc*=xc;}
   stiff_end();
   return y*x; 
}

#endif


REAL& REAL::operator = (const REAL& y) {
  MP_duplicate_wo_init(y.value,value);
  error=y.error;
  vsize=y.vsize;
  return (*this);
};

LAZY_BOOLEAN operator < (const REAL& x, const REAL& y) {
  REAL z=y-x;
  sizetype s;
  MP_getsize(z.value,s);
  if (sizetype_less(s,z.error) )
	{
	DEBUG5(1,"insufficient precisions %d*2^(%d) and %d*2^(%d) in comparing\n",
              x.error.mantissa,x.error.exponent,
              y.error.mantissa,y.error.exponent);
        return BOTTOM;
	}
  return  ((MP_sign((z.value))==1));
};

LAZY_BOOLEAN operator <  (const REAL&	 x, const int	   y){ return  (x<REAL(y)); };
LAZY_BOOLEAN operator <  (const REAL&	 x, const double   y){ return  (x<REAL(y)); };
LAZY_BOOLEAN operator <  (const REAL&	 x, const INTEGER& y){ return  (x<REAL(y)); };
LAZY_BOOLEAN operator <  (const REAL&	 x, const DYADIC&  y){ return  (x<REAL(y)); };
LAZY_BOOLEAN operator <  (const int	 x, const REAL&    y){ return  (REAL(x)<y); };
LAZY_BOOLEAN operator <  (const double   x, const REAL&    y){ return  (REAL(x)<y); };
LAZY_BOOLEAN operator <  (const INTEGER& x, const REAL&    y){ return  (REAL(x)<y); };
LAZY_BOOLEAN operator <  (const DYADIC&  x, const REAL&    y){ return  (REAL(x)<y); };

LAZY_BOOLEAN operator <= (const REAL&    x, const REAL&    y){ return  (x<y); };
LAZY_BOOLEAN operator <= (const REAL&	 x, const int	   y){ return  (x<=REAL(y)); };
LAZY_BOOLEAN operator <= (const REAL&	 x, const double   y){ return  (x<=REAL(y)); };
LAZY_BOOLEAN operator <= (const REAL&	 x, const INTEGER& y){ return  (x<=REAL(y)); };
LAZY_BOOLEAN operator <= (const REAL&	 x, const DYADIC&  y){ return  (x<=REAL(y)); };
LAZY_BOOLEAN operator <= (const int	 x, const REAL&    y){ return  (REAL(x)<=y); };
LAZY_BOOLEAN operator <= (const double   x, const REAL&    y){ return  (REAL(x)<=y); };
LAZY_BOOLEAN operator <= (const INTEGER& x, const REAL&    y){ return  (REAL(x)<=y); };
LAZY_BOOLEAN operator <= (const DYADIC&  x, const REAL&    y){ return  (REAL(x)<=y); };

LAZY_BOOLEAN operator >  (const REAL&    x, const REAL&    y){ return  (y<x); };
LAZY_BOOLEAN operator >  (const REAL&	 x, const int	   y){ return  (x>REAL(y)); };
LAZY_BOOLEAN operator >  (const REAL&	 x, const double   y){ return  (x>REAL(y)); };
LAZY_BOOLEAN operator >  (const REAL&	 x, const INTEGER& y){ return  (x>REAL(y)); };
LAZY_BOOLEAN operator >  (const REAL&	 x, const DYADIC&  y){ return  (x>REAL(y)); };
LAZY_BOOLEAN operator >  (const int	 x, const REAL&    y){ return  (REAL(x)>y); };
LAZY_BOOLEAN operator >  (const double   x, const REAL&    y){ return  (REAL(x)>y); };
LAZY_BOOLEAN operator >  (const INTEGER& x, const REAL&    y){ return  (REAL(x)>y); };
LAZY_BOOLEAN operator >  (const DYADIC&  x, const REAL&    y){ return  (REAL(x)>y); };

LAZY_BOOLEAN operator >= (const REAL&    x, const REAL&    y){ return  (y<x); };
LAZY_BOOLEAN operator >= (const REAL&	 x, const int	   y){ return  (x>=REAL(y)); };
LAZY_BOOLEAN operator >= (const REAL&	 x, const double   y){ return  (x>=REAL(y)); };
LAZY_BOOLEAN operator >= (const REAL&	 x, const INTEGER& y){ return  (x>=REAL(y)); };
LAZY_BOOLEAN operator >= (const REAL&	 x, const DYADIC&  y){ return  (x>=REAL(y)); };
LAZY_BOOLEAN operator >= (const int	 x, const REAL&    y){ return  (REAL(x)>=y); };
LAZY_BOOLEAN operator >= (const double   x, const REAL&    y){ return  (REAL(x)>=y); };
LAZY_BOOLEAN operator >= (const INTEGER& x, const REAL&    y){ return  (REAL(x)>=y); };
LAZY_BOOLEAN operator >= (const DYADIC&  x, const REAL&    y){ return  (REAL(x)>=y); };

LAZY_BOOLEAN operator == (const REAL&    x, const REAL&    y){ return  (y<x)&&(x<y) ; };
LAZY_BOOLEAN operator == (const REAL&	 x, const int	   y){ return  (x==REAL(y)); };
LAZY_BOOLEAN operator == (const REAL&	 x, const double   y){ return  (x==REAL(y)); };
LAZY_BOOLEAN operator == (const REAL&	 x, const INTEGER& y){ return  (x==REAL(y)); };
LAZY_BOOLEAN operator == (const REAL&	 x, const DYADIC&  y){ return  (x==REAL(y)); };
LAZY_BOOLEAN operator == (const int	 x, const REAL&    y){ return  (REAL(x)==y); };
LAZY_BOOLEAN operator == (const double   x, const REAL&    y){ return  (REAL(x)==y); };
LAZY_BOOLEAN operator == (const INTEGER& x, const REAL&    y){ return  (REAL(x)==y); };
LAZY_BOOLEAN operator == (const DYADIC&  x, const REAL&    y){ return  (REAL(x)==y); };

LAZY_BOOLEAN operator != (const REAL&    x, const REAL&    y){ return  (y<x)||(x<y) ;};
LAZY_BOOLEAN operator != (const REAL&	 x, const int	   y){ return  (x!=REAL(y)); };
LAZY_BOOLEAN operator != (const REAL&	 x, const double   y){ return  (x!=REAL(y)); };
LAZY_BOOLEAN operator != (const REAL&	 x, const INTEGER& y){ return  (x!=REAL(y)); };
LAZY_BOOLEAN operator != (const REAL&	 x, const DYADIC&  y){ return  (x!=REAL(y)); };
LAZY_BOOLEAN operator != (const int	 x, const REAL&    y){ return  (REAL(x)!=y); };
LAZY_BOOLEAN operator != (const double   x, const REAL&    y){ return  (REAL(x)!=y); };
LAZY_BOOLEAN operator != (const INTEGER& x, const REAL&    y){ return  (REAL(x)!=y); };
LAZY_BOOLEAN operator != (const DYADIC&  x, const REAL&    y){ return  (REAL(x)!=y); };

LAZY_BOOLEAN positive (const REAL& x, int k)
{ bool erg;
  sizetype ksize;
  sizetype_set(ksize,1,k);
  if (sizetype_less(ksize,x.error) && sizetype_less(x.vsize,x.error) ) 
    {
      DEBUG3(1,"insufficient precision %d*2^(%d) in test on positive\n",
		x.error.mantissa,x.error.exponent);
      return BOTTOM;
    }
  erg=(MP_sign(x.value)==1);
  return erg; 
};

DYADIC approx (const REAL& x, const int p) {
  MP_type result;
  MP_type erg;
  if ((ACTUAL_STACK.inlimit==0) && cache_mp.get(result)){
    MP_duplicate_w_init(result,erg);
    return DYADIC(erg);
  }

  sizetype psize;
  sizetype_set(psize,1,p+1);
  if (sizetype_less(psize,x.error) )
     {
       DEBUG4(1,"insufficient precision %d*2^(%d) in approx(%d)\n",
           x.error.mantissa,x.error.exponent,p);
       REITERATE(p-x.error.exponent);
     }
  MP_init(erg);
  MP_copy(x.value,erg,p-1);
  
  if ( ACTUAL_STACK.inlimit==0 ) {
    MP_duplicate_w_init(erg,result);
    cache_mp.put(result);
  }
  return DYADIC(erg); 
};

int size (const REAL& x) {
  int result=0;
  if ((ACTUAL_STACK.inlimit==0) && cache_i.get(result)) return result;
  sizetype xsize=x.vsize;
  sizetype ergsize;
  sizetype_add(ergsize,xsize,x.error);
  unsigned int value = ergsize.mantissa;
  for (unsigned int m=32>>1; m>0; m=m>>1 ) {
    if (value>= (1<<m) ){
      result+=m; value=value>>m;
    }
  }
  if ( (1<<result) != ergsize.mantissa) result+=1;
  result+=ergsize.exponent;

  if ( result < MP_min) throw iRRAM_Numerical_Exception(iRRAM_underflow_error);

  sizetype_set(xsize,1,result-2);
  sizetype_inc(xsize,x.error);
  if (sizetype_less(x.vsize,xsize) )
	{  
	DEBUG5(1,"insufficient precision %d*2^(%d) in size %d*2^(%d)\n",
           x.error.mantissa,x.error.exponent,x.vsize.mantissa,x.vsize.exponent);
	REITERATE(0);
	}

  if ( ACTUAL_STACK.inlimit==0 ) cache_i.put(result);
  return result;
};


#ifdef MP_double_to_mp 
REAL::REAL(double d){
  if ( !isfinite(d) ) throw iRRAM_Numerical_Exception(iRRAM_conversion_from_infinite);
  MP_init(value); 
  MP_double_to_mp(d,value);
  sizetype_exact(error);
  MP_getsize(value,vsize);  
}
#else
REAL::REAL(double d){
  if ( !isfinite(d) ) throw iRRAM_Numerical_Exception(iRRAM_conversion_from_infinite);
   REAL result =0;
   int sign=0;
   int exp=0;
   int v;
   if ( d!= 0.0 ) {
     if  (d< 0) {sign=-1;d=-d;}
     else sign=1;
     while ( d >1024.0) {d=d/1024; exp+=10;}
     while ( d <1.0) {d=d*1024; exp-=10;}
     while ( d >0.0) {
       exp-=20;
       v=(int)(d*1048576);
       result=result*1048576+v;
       d=d*(1024*1024)-v;
     }
     if (exp < 0) {
       while ( exp<=-20 ) {exp+=20;result=result/(1048576);}
       while ( exp<=-10 ) {exp+=10;result=result/(1024);}
     } else {
       while ( exp>=20 ) {exp-=20;result=result*(1048576);}
       while ( exp>=10 ) {exp-=10;result=result*(1024);}
     }
   }
   result=sign*result;
   error=result.error;
   MP_duplicate_w_init(result.value,value);
   MP_getsize(value,size);
}
#endif


REAL abs (const REAL& x){
  MP_type zvalue;
  MP_init(zvalue);
  MP_abs(x.value,zvalue);
  return REAL(zvalue,x.error);
}

REAL round2(const REAL& x) {
  return REAL(INTEGER(x));
};

int round (const REAL& x) {
  return int(INTEGER(x));
};

int upperbound (const REAL& x) {
  int result;
  sizetype ergsize;
  if ( (ACTUAL_STACK.inlimit==0) && cache_i.get(result)) return result;
  sizetype_add(ergsize,x.vsize,x.error);
  while (ergsize.mantissa>1) {
    ergsize.mantissa=ergsize.mantissa>>1;
    ergsize.exponent+=1;
  }
  result=ergsize.exponent;
  if (ACTUAL_STACK.inlimit==0 ) cache_i.put(result);
  return result;
};

LAZY_BOOLEAN bound (const REAL& x, const int k) {
  sizetype lowsize,ksize,highsize;
  int erg;
  sizetype_set(ksize,1,k);
  sizetype_set(lowsize,1,k-1);
  sizetype_inc(lowsize,x.error);
  sizetype_add(highsize,x.vsize,x.error);
  if (sizetype_less(x.vsize,lowsize) && sizetype_less(ksize,highsize) ) 
    {  
      DEBUG6(1,"insufficient precision %d*2^(%d) in bounding by 2^(%d) for argument of size  %d*2^(%d)\n",
              x.error.mantissa,x.error.exponent,k,x.vsize.mantissa,x.vsize.exponent);
      return BOTTOM;
    }
  erg=(sizetype_less(x.vsize,ksize) );
  return erg;
};



#define ABSOLUTE 0
#define REALTIVE 1
void precision_policy (int policy){ACTUAL_STACK.prec_policy=policy;};
void stiff_begin (){ ACTUAL_STACK.actual_prec += ACTUAL_STACK.prec_inc;}
void stiff_end   (){ ACTUAL_STACK.actual_prec -= ACTUAL_STACK.prec_inc;}
void continous_begin (){ACTUAL_STACK.inlimit+=1;};
void continous_end   (){ACTUAL_STACK.inlimit-=1;};

string swrite(const REAL& x, const int w, const int form)
{
  string result;
  if ( ACTUAL_STACK.inlimit==0 && cache_s.get(result)) return result;

  int width=w;
  if (width<9) width=9;
  char* erg;

  if (form==iRRAM_float_absolute) {
    sizetype psize; 
    int p=-10*(width-8)/3;
    sizetype_set(psize,1,p);
    int s=MP_size(x.value);
    int mantissa=(int)((s-x.error.exponent-GUARD_BITS)*.30103); 
    if (sizetype_less(psize,x.error)|| (s>p && mantissa+8 < width) ) {
      DEBUG4(1,"insufficient precision %d*2^(%d) in conversion with precision 2^(%d)\n",
         x.error.mantissa,x.error.exponent,p);
      REITERATE(p-x.error.exponent);
    }

    if (s<=p) {
      erg=new char[width+1];
      strcpy(erg,"  0 ");
      for (int j=4;j<width;j++) erg[j]=' ';
      erg[width]=0;
    } else {
      erg=MP_swrite(x.value,width);
    }

  } else if (form==iRRAM_float_relative){
     continous_begin();
     int p=size(x)-2-10*(width-8)/3; 
//We have to prevent that "size" changes the multi-value-cache here! 
//Otherwise, any iteration would not use the same amount of 
// cache as the first run!
     continous_end();

     sizetype psize;
     sizetype_set(psize,1,p);
     int s=MP_size(x.value);
     int mantissa=(int)((s-x.error.exponent-GUARD_BITS)*.30103); 
     if (sizetype_less(psize,x.error)|| (s>p && mantissa+8 < width) ) {
       DEBUG4(1,"insufficient precision %d*2^(%d) in writing with precision 2^(%d)\n",
         x.error.mantissa,x.error.exponent,p);
       REITERATE(p-x.error.exponent);
     }
     erg=MP_swrite(x.value,width);
  
  
  } else if (form==iRRAM_float_show){

    int mantissa=(int)(int((MP_size(x.value)-x.error.exponent-GUARD_BITS))*.30103); 
    if (mantissa < 1 ) { 
      erg=new char[width+1];
      sprintf(erg," .*E%05.0f",1+(x.error.exponent+GUARD_BITS)*.30103);
      for (int j=9;j<width;j++) erg[j]=' ';
    } else { 
      if ( mantissa>width-8 ) mantissa=width-8;
      char * erg2=MP_swrite(x.value,mantissa+8);
      erg=new char[width+1];
      strncpy(erg,erg2,mantissa+8);
      for (int j=mantissa+8;j<width;j++) erg[j]=' ';
      erg[width]=0;
    }

  }

 
  result=erg;
  free(erg);
  if ( ACTUAL_STACK.inlimit==0 ) cache_s.put(result);
  return result;
}


void   rwrite  (const REAL& x, const int w){ rout << swrite(x,w,iRRAM_float_absolute); }
void   rwritee (const REAL& x, const int w){ rout << swrite(x,w,iRRAM_float_relative); }
void rshow(const REAL& x, const int w){ rout << swrite(x,w,iRRAM_float_show); }


void REAL::rcheck(int n) const
{
cerr << "Value: ";
char* c=MP_swrite(value,n);
cerr << c;
free(c);
cerr << ", Size:  "<<vsize.mantissa<<"*2^("<<vsize.exponent<<")\n\n";
cerr << ", Error: "<<error.mantissa<<"*2^("<<error.exponent<<")\n\n";
}

void REAL::adderror (sizetype error)
{ sizetype_inc((*this).error,error);
  ADJUST((*this).error);
}

void REAL::seterror (sizetype error)
{ (*this).error=error;
}

void REAL::geterror (sizetype& error) const 
{ error=(*this).error;
}

